@@ -1,3 +1,4 @@
+import math
 from typing import Optional
 
 from .transformer import TransformerEncoder
@@ -11,7 +12,7 @@ class SequencePool(nn.Module):
     """
     As described in [Hasani et al. (2021) *''Escaping the Big Data Paradigm with
         Compact Transformers''*](https://arxiv.org/abs/2104.05704). It can be viewed
-        as. ageneralisation of average pooling.
+        as a generalisation of average pooling.
     """
 
     def __init__(self, d_model, linear_module, out_dim):
@@ -54,6 +55,7 @@ class CCTEncoder(nn.Module):
         conv_pooling_kernel_size=3,
         conv_pooling_kernel_stride=2,
         conv_pooling_kernel_padding=1,
+        transformer_position_embedding="absolute",  # absolute or relative
         transformer_embedding_size=256,
         transformer_layers=7,
         transformer_heads=4,
@@ -84,22 +86,26 @@ class CCTEncoder(nn.Module):
 
         self.image_size = image_size
 
-        output_size = (
+        # XXX: We assume a square image here
+        output_size = math.floor(
             (image_size + 2 * conv_pooling_kernel_padding - conv_pooling_kernel_size)
             / conv_pooling_kernel_stride
-        ) + 1
+            + 1
+        )  # output of pooling
 
-        self.sequence_length = int(output_size) ** 2
+        self.sequence_length = output_size**2
 
         if conv_pooling_type == "maxpool":
 
             conv_out_channels = transformer_embedding_size
+            pooling_out_channels = transformer_embedding_size
+            self.pooling_adapter = nn.Identity()
 
             self.pool = nn.Sequential(
                 *[
-                    Rearrange(
+                    Rearrange(  # rearrange in case we're using XGLU activation
                         "N C H W -> N H W C"
-                    ),  # rearrange in case we're using XGLU activation
+                    ),
                     self.activation,
                     Rearrange("N H W C -> N C H W"),
                     nn.MaxPool2d(
@@ -107,24 +113,35 @@ class CCTEncoder(nn.Module):
                         stride=conv_pooling_kernel_stride,
                         padding=conv_pooling_kernel_padding,
                     ),
-                    Rearrange("N C H W -> N (H W) C"),
                 ]
             )
-        elif conv_pooling_type == "concat":
 
+        elif conv_pooling_type == "concat":
+            # TODO: make all of this adapter business part of concatpool
             conv_out_channels = int(
                 round(transformer_embedding_size / (conv_pooling_kernel_size**2))
             )
+            pooling_out_channels = conv_pooling_kernel_size**2 * conv_out_channels
+            pooling_adapter_out_channels = transformer_embedding_size
+            if activation.__name__.endswith("GLU"):
+                pooling_adapter_out_channels *= 2
+            self.pooling_adapter = nn.Sequential(
+                *[
+                    Rearrange("N C H W -> N (H W) C"),
+                    nn.Linear(pooling_out_channels, pooling_adapter_out_channels),
+                    self.activation,
+                ]
+            )
 
-            self.pool = ConcatPool(
-                conv_out_channels,
-                conv_pooling_kernel_size,
-                conv_pooling_kernel_stride,
-                conv_pooling_kernel_padding,
-                transformer_embedding_size,
-                activation,
-                activation_kwargs,
-                linear_module,
+            self.pool = nn.Sequential(
+                *[
+                    ConcatPool(
+                        conv_pooling_kernel_size,
+                        stride=conv_pooling_kernel_stride,
+                        padding=conv_pooling_kernel_padding,
+                    ),
+                    self.pooling_adapter,
+                ]
             )
 
         if transformer_layers > 0:
@@ -133,6 +150,8 @@ class CCTEncoder(nn.Module):
                 transformer_embedding_size,
                 transformer_layers,
                 transformer_heads,
+                position_embedding_type=transformer_position_embedding,
+                source_size=(output_size, output_size),
                 mlp_ratio=transformer_mlp_ratio,
                 activation=activation,
                 activation_kwargs=activation_kwargs,
@@ -164,7 +183,13 @@ class CCTEncoder(nn.Module):
             linear_module=linear_module,
         )
 
-        self.encoder = nn.Sequential(*[self.conv, self.pool, self.transformer])
+        self.encoder = nn.Sequential(
+            *[
+                self.conv,
+                self.pool,
+                self.transformer,
+            ]
+        )
 
     def forward(self, x):
         return self.encoder(x)
@@ -186,6 +211,7 @@ class CCT(nn.Module):
         pooling_kernel_size=3,
         pooling_kernel_stride=2,
         pooling_kernel_padding=1,
+        transformer_position_embedding="absolute",  # absolute or relative
         transformer_embedding_size=256,
         transformer_layers=7,
         transformer_heads=4,
@@ -215,6 +241,7 @@ class CCT(nn.Module):
             conv_pooling_kernel_size=pooling_kernel_size,
             conv_pooling_kernel_stride=pooling_kernel_stride,
             conv_pooling_kernel_padding=pooling_kernel_padding,
+            transformer_position_embedding=transformer_position_embedding,
             transformer_embedding_size=transformer_embedding_size,
             transformer_layers=transformer_layers,
             transformer_heads=transformer_heads,
